#include <stdio.h>
#include <string.h>

struct nlist { /* table entry: */
    struct nlist *next; /* next entry in chain */
    double name;
    int defn;
};

#define HASHSIZE 191
static struct nlist *hashtab[HASHSIZE]; /* pointer table */

/* hash: form hash value for string s */
static unsigned hash(double v)
{
    char *s = (char*)&v;
    unsigned hashval;
    int i;
    for (i = 0, hashval = 0; i < 4; i++, s++)
      hashval = *s + 31 * hashval;
    return hashval % HASHSIZE;
}

/* lookup: look for s in hashtab */
struct nlist *lookup(double s)
{
    struct nlist *np;
    for (np = hashtab[hash(s)]; np != NULL; np = np->next)
        if (s == np->name)
          return np; /* found */
    return NULL; /* not found */
}

/* install: put (name, defn) in hashtab */
struct nlist *install(double name)
{
    struct nlist *np;
    unsigned hashval;
    if ((np = lookup(name)) == NULL) { /* not found */
        np = (struct nlist *) malloc(sizeof(*np));
        if (np == NULL)
          return NULL;
        np->name = name;
        np->defn = 0;
        hashval = hash(name);
        np->next = hashtab[hashval];
        hashtab[hashval] = np;
    } /* already there */
    np->defn++;
    return np;
}

/* keys */
void keys(void fn(double, int))
{
    int i;
    for (i = 0; i < HASHSIZE; i++)
    {
        struct nlist *np = hashtab[i];
        if (np != NULL)
        {
            struct nlist *np2;
            for (np2 = np; np2 != NULL; np2 = np2->next)
            {
                fn(np2->name, np2->defn);
            }
        }
    }
}

